package org.processmining.dataawarecnetminer.model;

import java.util.Collection;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.dataawarecnetminer.parallel.Command;
import org.processmining.dataawarecnetminer.parallel.TaskExecutor;
import org.processmining.log.utils.XUtils;
import org.processmining.models.causalgraph.Relation;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multiset;

public interface EventRelationStorage {
	
	public static final String ARTIFICIAL_START = "ARTIFICIAL_START";
	public static final String ARTIFICIAL_END = "ARTIFICIAL_END";

	public class Factory {

		public static EventRelationStorage createEventRelations(XEventClasses eventClasses) {
			return new EventRelationStorageImpl(eventClasses);
		}

		public static EventRelationStorage createEventRelations(Iterable<XTrace> log, XEventClassifier classifier,
				ExecutorService executor) {
			return createEventRelations(log, XUtils.createEventClasses(classifier, log), executor);
		}

		public static EventRelationStorage createEventRelations(Iterable<XTrace> log, XEventClasses eventClasses,
				ExecutorService executor) {
			final EventRelationStorage relationStorage = createEventRelations(eventClasses);

			try {
				TaskExecutor.Factory.getInstance().execute(executor, log, new Command<XTrace>() {

					public void execute(XTrace trace) {
						relationStorage.addTrace(trace);
					}
				});
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
			}

			return relationStorage;
		}

		public static EventRelationStorage createEventRelations(Iterable<XTrace> log, XEventClassifier classifier) {
			ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
			try {
				return createEventRelations(log, classifier, executor);
			} finally {
				executor.shutdown();
			}

		}

	}

	// Event classes
	ImmutableSet<XEventClass> getEventClasses();

	ImmutableMap<String, XEventClass> getEventClassesById();

	XEventClassifier getClassifier();

	XEventClass getEventClass(XEvent event);

	// Artificial start and end event class
	XEventClass getStartEventClass();

	XEventClass getEndEventClass();

	boolean includesStartOrEnd(Relation relation);

	// Convenience methods for relations
	Iterable<Relation> allNonEqualCombinations();

	Iterable<Relation> allCombinations();
	
	Iterable<Relation> allObservedCombinations();

	Iterable<Relation> allCombinations(XEventClass source);

	// Mutation methods
	void addTrace(XTrace trace);

	void addEventuallyFollows(XEventClass source, XEventClass target);

	void addLengthTwoFollows(XEventClass source, XEventClass target);

	void addDirectlyFollows(XEventClass source, XEventClass target);

	// Query methods
	Collection<XTrace> getAllTraces();

	Multiset<Relation> getDirectlyFollowsRelations();
	
	Set<XEventClass> directlyFollows(XEventClass source);

	Multiset<Relation> getEventuallyFollowsRelations();
	
	Multiset<XEventClass> getEventClassOccurence();	

	// Measures & counts
	int countTraces();

	int countEvents();

	int countOccurence(XEventClass eventClass);

	int countDirectlyFollows(XEventClass source, XEventClass target);

	int countDirectlyFollows(Relation relation);

	int countLengthTwoFollows(XEventClass source, XEventClass target);

	int countEventuallyFollows(XEventClass source, XEventClass target);

	DependencyRelation getDependencyRelation(Relation relation);

	DependencyRelation getL1LoopRelation(Relation relation);

	DependencyRelation getL2LoopRelation(Relation relation);

	DependencyRelation getLongDistanceDependency(Relation element);

}